/* version.h - 22:16 GMT +10:00 Fri 20 Aug 1993 - Geoffrey Tobin. */

#ifndef VERSION_H
#define VERSION_H

#define dvgt_version  \
"This is dvgt 3.51L3 (DVItoVDU, v. 3.0, gt mod 5 rev 1 Linux 3) - 10 Jul 95"

#endif /* VERSION_H */

/* end version.h */
